NAME = "traceml"
VERSION = "1.2.1-rc2"
DESC = (
    "Engine for ML/Data tracking, visualization, dashboards, and model UI for Polyaxon."
)
URL = "https://github.com/polyaxon/traceml"
AUTHOR = "Polyaxon, Inc."
EMAIL = "contact@polyaxon.com"
LICENSE = "Apache 2.0"
